/*
 * Copyright (C) 2003 Maxim Stepin ( maxst@hiend3d.com )
 *
 * Copyright (C) 2010 Cameron Zemek ( grom@zeminvaders.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdint.h>
#include <alloca.h>

#if ( \
    defined(__ARM_ARCH_6__) || \
    defined(__ARM_ARCH_6J__) || \
    defined(__ARM_ARCH_6K__) || \
    defined(__ARM_ARCH_6Z__) || \
    defined(__ARM_ARCH_6ZK__) || \
    defined(__ARM_ARCH_6T2__) || \
    defined(__ARM_ARCH_7__) || \
    defined(__ARM_ARCH_7A__) || \
    defined(__ARM_ARCH_7R__) || \
    defined(__ARM_ARCH_7M__) || \
    defined(__ARM_ARCH_7S__) || \
    (defined(_M_ARM) && (_M_ARM >= 6)) || \
    (defined(__TARGET_ARCH_ARM) && (__TARGET_ARCH_ARM >= 6)) || \
    (defined(__TARGET_ARCH_THUMB) && (__TARGET_ARCH_THUMB >= 3)) \
)
    #define ARMV6 1
#else
    #undef ARMV6
#endif

#if ( \
    defined(__ARM_ARCH_7__) || \
    defined(__ARM_ARCH_7A__) || \
    defined(__ARM_ARCH_7R__) || \
    defined(__ARM_ARCH_7M__) || \
    defined(__ARM_ARCH_7S__) || \
    (defined(_M_ARM) && (_M_ARM >= 7)) || \
    (defined(__TARGET_ARCH_ARM) && (__TARGET_ARCH_ARM >= 7)) || \
    (defined(__TARGET_ARCH_THUMB) && (__TARGET_ARCH_THUMB >= 4)) \
)
    #define ARMV7 1
#else
    #undef ARMV7
#endif

#define ALIGNMENT 8



#define Ymask 0x00FF0000
#define Umask 0x0000FF00
#define Vmask 0x000000FF
#define trY   0x00300000
#define trU   0x00000700
#define trV   0x00000006

/* RGB to YUV lookup table */
//extern uint32_t   RGBtoYUV[65536];

static inline int yuv_diff(uint32_t yuv1, uint32_t yuv2)
{
#if defined(ARMV6)
    register uint32_t tmp1, tmp2;

    asm(
        "usub8 %[tmp1], %[value1], %[value2]   \n\t" // tmp1 = value1 - value2             // tmp1 = yuv1 - yuv2
        "usub8 %[value2], %[value2], %[value1] \n\t" // value2 = value2 - value1           // value2 = yuv2 - yuv1
        "mov %[tmp2], #0                       \n\t" // tmp2 = 0
        "movw %[value1], #0x0706               \n\t"
        "movt %[value1], #0x30                 \n\t" // value1 = 0x300706
        "sel %[tmp1], %[value2], %[tmp1]       \n\t" // tmp1 = (value2 >= 0)?value2:tmp1   // tmp1 = abs(yuv1 - yuv2)
        "usub8 %[value2], %[value1], %[tmp1]   \n\t" // value2 = value1 - tmp1             // value2 = 0x300706 - abs(yuv1 - yuv2)
        "sel %[value1], %[tmp2], %[value1]         " // value1 = (value2 >= 0)?tmp2:value1 // value1 = (0x300706 >= abs(yuv1 - yuv2))?0:0x300706

        : [value1] "+r" (yuv1), [value2] "+r" (yuv2), [tmp1] "=r" (tmp1), [tmp2] "=r" (tmp2)
        :
        : "cc"
    );

    return yuv1;
#else
    return ( ( abs((yuv1 & Ymask) - (yuv2 & Ymask)) > trY ) ||
            ( abs((yuv1 & Umask) - (yuv2 & Umask)) > trU ) ||
            ( abs((yuv1 & Vmask) - (yuv2 & Vmask)) > trV ) );
#endif
}

/* Test if there is difference in color */
/*static inline int Diff(uint32_t w1, uint32_t w2)
{
    return yuv_diff(RGBtoYUV[w1], RGBtoYUV[w2]);
}*/


#if defined(ARMV7)
void convert_yuv_16(const uint16_t *src, uint32_t *dst, int width);
void convert_yuv_rgb_8(const uint8_t *src, uint32_t *yuvdst, uint16_t *rgbdst, const uint16_t *pal, int width);
#else
static void convert_yuv_16(const uint16_t *src, uint32_t *dst, int width)
{
    int index;

    for (index = 0; index < width; index++)
    {
        uint32_t r, g, b;
        uint32_t y, u, v;

        r = (src[index] << 3) & (31 << 3);
        g = (src[index] >> 3) & (63 << 2);
        b = (src[index] >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        // this is the original HQX formula - using binary floating point numbers
        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        // this is the same formula, but using more precise coeficients and decimal floating point numbers
        // it gives slightly different results, but it's more precise
        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        // this is the same formula, but using 9.23 fixed point numbers
        // it gives the results as when using using decimal floating point numbers
        //y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        ////u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        //u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        //v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this is the same formula, but minimizing coeficients and fraction precision
        // it gives the results as when using using decimal floating point numbers
        //
        // this(>>18) works for rgb888 values created from rgb565
        y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        dst[index] = (y << 16) + (u << 8) + v;

        //dst[index] = RGBtoYUV[src[index]];
    }

    dst[-1] = dst[0];
    dst[width] = dst[width - 1];
}

static void convert_yuv_rgb_8(const uint8_t *src, uint32_t *yuvdst, uint16_t *rgbdst, const uint16_t *pal, int width)
{
    int index;

    for (index = 0; index < width; index++)
    {
        uint32_t palvalue;
        uint32_t r, g, b;
        uint32_t y, u, v;

        palvalue = pal[src[index]];
        rgbdst[index] = palvalue;

        r = (palvalue << 3) & (31 << 3);
        g = (palvalue >> 3) & (63 << 2);
        b = (palvalue >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        // this is the original HQX formula - using binary floating point numbers
        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        // this is the same formula, but using more precise coeficients and decimal floating point numbers
        // it gives slightly different results, but it's more precise
        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        // this is the same formula, but using 9.23 fixed point numbers
        // it gives the results as when using using decimal floating point numbers
        //y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        ////u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        //u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        //v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this is the same formula, but minimizing coeficients and fraction precision
        // it gives the results as when using using decimal floating point numbers
        //
        // this(>>18) works for rgb888 values created from rgb565
        y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        yuvdst[index] = (y << 16) + (u << 8) + v;
    }

    yuvdst[-1] = yuvdst[0];
    yuvdst[width] = yuvdst[width - 1];
}
#endif

static void calculate_pattern(const uint32_t *yuvsrc1, const uint32_t *yuvsrc2, const uint32_t *yuvsrc3, uint8_t *dst, int width)
{
    uint32_t e, newb, newe, newh;
    uint8_t pattern;

    // before first
    {
        newb = yuvsrc1[-1];
        newe = yuvsrc2[-1];
        newh = yuvsrc3[-1];
        pattern = 0;
        if (yuv_diff(newe, newb    )) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(newe, newh    )) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        dst[-1] = pattern;
    }

    for (; width != 0; width--)
    {
        e = newe;
        pattern = 0;
        if (yuv_diff(e, yuvsrc1[-1])) pattern |= 1 << 0;
        if (yuv_diff(e, newb       )) pattern |= 1 << 1;
        newb = yuvsrc1[ 1];
        if (yuv_diff(e, newb       )) pattern |= 1 << 2;
        if (yuv_diff(e, yuvsrc2[-1])) pattern |= 1 << 3;
        newe = yuvsrc2[ 1];
        if (yuv_diff(e, newe       )) pattern |= 1 << 4;
        if (yuv_diff(e, yuvsrc3[-1])) pattern |= 1 << 5;
        if (yuv_diff(e, newh       )) pattern |= 1 << 6;
        newh = yuvsrc3[ 1];
        if (yuv_diff(e, newh       )) pattern |= 1 << 7;

        *dst = pattern;

        yuvsrc1++;
        yuvsrc2++;
        yuvsrc3++;
        dst++;
    }

    // after last
    {
        pattern = 0;
        if (yuv_diff(newe, newb    )) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(newe, newh    )) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        *dst = pattern;
    }
}


/* Interpolate functions */
#define GROW(n) n = (n | n << 16) & 0x07e0f81f
static uint16_t inline pack(uint32_t n)
{
    n &= 0x07e0f81f;
    return n | (n >> 16);
}

static uint16_t inline blend1(uint32_t A, uint32_t B)
{
    GROW(A); GROW(B);
    A = (A * 3 + B) >> 2;
    return pack(A);
}

static uint16_t inline blend2(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 2 + B + C) >> 2);
}

static uint16_t inline blend3(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 5 + B * 2 + C) >> 3);
}

static uint16_t inline blend4(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 6 + B + C) >> 3);
}

static uint16_t inline blend5(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 2 + (B + C) * 3) >> 3);
}

static uint16_t inline blend6(uint32_t A, uint32_t B, uint32_t C)
{
    GROW(A); GROW(B); GROW(C);
    return pack((A * 14 + B + C) >> 4);
}


#define PIXEL00_0     *dst = w[5];
#define PIXEL00_10    *dst = blend1(w[5], w[1]);
#define PIXEL00_11    *dst = blend1(w[5], w[4]);
#define PIXEL00_12    *dst = blend1(w[5], w[2]);
#define PIXEL00_20    *dst = blend2(w[5], w[4], w[2]);
#define PIXEL00_21    *dst = blend2(w[5], w[1], w[2]);
#define PIXEL00_22    *dst = blend2(w[5], w[1], w[4]);
#define PIXEL00_60    *dst = blend3(w[5], w[2], w[4]);
#define PIXEL00_61    *dst = blend3(w[5], w[4], w[2]);
#define PIXEL00_70    *dst = blend4(w[5], w[4], w[2]);
#define PIXEL00_90    *dst = blend5(w[5], w[4], w[2]);
#define PIXEL00_100   *dst = blend6(w[5], w[4], w[2]);
#define PIXEL01_0     *(dst+1) = w[5];
#define PIXEL01_10    *(dst+1) = blend1(w[5], w[3]);
#define PIXEL01_11    *(dst+1) = blend1(w[5], w[2]);
#define PIXEL01_12    *(dst+1) = blend1(w[5], w[6]);
#define PIXEL01_20    *(dst+1) = blend2(w[5], w[2], w[6]);
#define PIXEL01_21    *(dst+1) = blend2(w[5], w[3], w[6]);
#define PIXEL01_22    *(dst+1) = blend2(w[5], w[3], w[2]);
#define PIXEL01_60    *(dst+1) = blend3(w[5], w[6], w[2]);
#define PIXEL01_61    *(dst+1) = blend3(w[5], w[2], w[6]);
#define PIXEL01_70    *(dst+1) = blend4(w[5], w[2], w[6]);
#define PIXEL01_90    *(dst+1) = blend5(w[5], w[2], w[6]);
#define PIXEL01_100   *(dst+1) = blend6(w[5], w[2], w[6]);
#define PIXEL10_0     *(dst+dpL) = w[5];
#define PIXEL10_10    *(dst+dpL) = blend1(w[5], w[7]);
#define PIXEL10_11    *(dst+dpL) = blend1(w[5], w[8]);
#define PIXEL10_12    *(dst+dpL) = blend1(w[5], w[4]);
#define PIXEL10_20    *(dst+dpL) = blend2(w[5], w[8], w[4]);
#define PIXEL10_21    *(dst+dpL) = blend2(w[5], w[7], w[4]);
#define PIXEL10_22    *(dst+dpL) = blend2(w[5], w[7], w[8]);
#define PIXEL10_60    *(dst+dpL) = blend3(w[5], w[4], w[8]);
#define PIXEL10_61    *(dst+dpL) = blend3(w[5], w[8], w[4]);
#define PIXEL10_70    *(dst+dpL) = blend4(w[5], w[8], w[4]);
#define PIXEL10_90    *(dst+dpL) = blend5(w[5], w[8], w[4]);
#define PIXEL10_100   *(dst+dpL) = blend6(w[5], w[8], w[4]);
#define PIXEL11_0     *(dst+dpL+1) = w[5];
#define PIXEL11_10    *(dst+dpL+1) = blend1(w[5], w[9]);
#define PIXEL11_11    *(dst+dpL+1) = blend1(w[5], w[6]);
#define PIXEL11_12    *(dst+dpL+1) = blend1(w[5], w[8]);
#define PIXEL11_20    *(dst+dpL+1) = blend2(w[5], w[6], w[8]);
#define PIXEL11_21    *(dst+dpL+1) = blend2(w[5], w[9], w[8]);
#define PIXEL11_22    *(dst+dpL+1) = blend2(w[5], w[9], w[6]);
#define PIXEL11_60    *(dst+dpL+1) = blend3(w[5], w[8], w[6]);
#define PIXEL11_61    *(dst+dpL+1) = blend3(w[5], w[6], w[8]);
#define PIXEL11_70    *(dst+dpL+1) = blend4(w[5], w[6], w[8]);
#define PIXEL11_90    *(dst+dpL+1) = blend5(w[5], w[6], w[8]);
#define PIXEL11_100   *(dst+dpL+1) = blend6(w[5], w[6], w[8]);

static void fill_dst(const uint16_t *rgbsrc1, const uint16_t *rgbsrc2, const uint16_t *rgbsrc3, uint16_t *dst, const uint8_t *patternTable, int width)
{
    int  i;
    uint32_t  w[10];
    int dpL = width * 2;

    //   +----+----+----+
    //   |    |    |    |
    //   | w1 | w2 | w3 |
    //   +----+----+----+
    //   |    |    |    |
    //   | w4 | w5 | w6 |
    //   +----+----+----+
    //   |    |    |    |
    //   | w7 | w8 | w9 |
    //   +----+----+----+

    w[2] = *rgbsrc1;
    w[5] = *rgbsrc2;
    w[8] = *rgbsrc3;

    for (i=0; i<width; i++)
    {
        w[1] = w[2];
        w[4] = w[5];
        w[7] = w[8];

        w[2] = *rgbsrc1;
        w[5] = *rgbsrc2;
        w[8] = *rgbsrc3;

        if (i<width-1)
        {
            w[3] = *(rgbsrc1 + 1);
            w[6] = *(rgbsrc2 + 1);
            w[9] = *(rgbsrc3 + 1);
        }
        else
        {
            w[3] = w[2];
            w[6] = w[5];
            w[9] = w[8];
        }

        switch (patternTable[i])
        {
            case 0:
            case 1:
            case 4:
            case 32:
            case 128:
            case 5:
            case 132:
            case 160:
            case 33:
            case 129:
            case 36:
            case 133:
            case 164:
            case 161:
            case 37:
            case 165:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 2:
            case 34:
            case 130:
            case 162:
                {
                    PIXEL00_22
                    PIXEL01_21
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 16:
            case 17:
            case 48:
            case 49:
                {
                    PIXEL00_20
                    PIXEL01_22
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 64:
            case 65:
            case 68:
            case 69:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_21
                    PIXEL11_22
                    break;
                }
            case 8:
            case 12:
            case 136:
            case 140:
                {
                    PIXEL00_21
                    PIXEL01_20
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 3:
            case 35:
            case 131:
            case 163:
                {
                    PIXEL00_11
                    PIXEL01_21
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 6:
            case 38:
            case 134:
            case 166:
                {
                    PIXEL00_22
                    PIXEL01_12
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 20:
            case 21:
            case 52:
            case 53:
                {
                    PIXEL00_20
                    PIXEL01_11
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 144:
            case 145:
            case 176:
            case 177:
                {
                    PIXEL00_20
                    PIXEL01_22
                    PIXEL10_20
                    PIXEL11_12
                    break;
                }
            case 192:
            case 193:
            case 196:
            case 197:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_21
                    PIXEL11_11
                    break;
                }
            case 96:
            case 97:
            case 100:
            case 101:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_12
                    PIXEL11_22
                    break;
                }
            case 40:
            case 44:
            case 168:
            case 172:
                {
                    PIXEL00_21
                    PIXEL01_20
                    PIXEL10_11
                    PIXEL11_20
                    break;
                }
            case 9:
            case 13:
            case 137:
            case 141:
                {
                    PIXEL00_12
                    PIXEL01_20
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 18:
            case 50:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 80:
            case 81:
                {
                    PIXEL00_20
                    PIXEL01_22
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 72:
            case 76:
                {
                    PIXEL00_21
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 10:
            case 138:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 66:
                {
                    PIXEL00_22
                    PIXEL01_21
                    PIXEL10_21
                    PIXEL11_22
                    break;
                }
            case 24:
                {
                    PIXEL00_21
                    PIXEL01_22
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 7:
            case 39:
            case 135:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 148:
            case 149:
            case 180:
                {
                    PIXEL00_20
                    PIXEL01_11
                    PIXEL10_20
                    PIXEL11_12
                    break;
                }
            case 224:
            case 228:
            case 225:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 41:
            case 169:
            case 45:
                {
                    PIXEL00_12
                    PIXEL01_20
                    PIXEL10_11
                    PIXEL11_20
                    break;
                }
            case 22:
            case 54:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 208:
            case 209:
                {
                    PIXEL00_20
                    PIXEL01_22
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 104:
            case 108:
                {
                    PIXEL00_21
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 11:
            case 139:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 19:
            case 51:
                {
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL00_11
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL00_60
                        PIXEL01_90
                    }
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 146:
            case 178:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                        PIXEL11_12
                    }
                    else
                    {
                        PIXEL01_90
                        PIXEL11_61
                    }
                    PIXEL10_20
                    break;
                }
            case 84:
            case 85:
                {
                    PIXEL00_20
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL01_11
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL01_60
                        PIXEL11_90
                    }
                    PIXEL10_21
                    break;
                }
            case 112:
            case 113:
                {
                    PIXEL00_20
                    PIXEL01_22
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL10_12
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL10_61
                        PIXEL11_90
                    }
                    break;
                }
            case 200:
            case 204:
                {
                    PIXEL00_21
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                        PIXEL11_11
                    }
                    else
                    {
                        PIXEL10_90
                        PIXEL11_60
                    }
                    break;
                }
            case 73:
            case 77:
                {
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL00_12
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL00_61
                        PIXEL10_90
                    }
                    PIXEL01_20
                    PIXEL11_22
                    break;
                }
            case 42:
            case 170:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                        PIXEL10_11
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL10_60
                    }
                    PIXEL01_21
                    PIXEL11_20
                    break;
                }
            case 14:
            case 142:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                        PIXEL01_12
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL01_61
                    }
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 67:
                {
                    PIXEL00_11
                    PIXEL01_21
                    PIXEL10_21
                    PIXEL11_22
                    break;
                }
            case 70:
                {
                    PIXEL00_22
                    PIXEL01_12
                    PIXEL10_21
                    PIXEL11_22
                    break;
                }
            case 28:
                {
                    PIXEL00_21
                    PIXEL01_11
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 152:
                {
                    PIXEL00_21
                    PIXEL01_22
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 194:
                {
                    PIXEL00_22
                    PIXEL01_21
                    PIXEL10_21
                    PIXEL11_11
                    break;
                }
            case 98:
                {
                    PIXEL00_22
                    PIXEL01_21
                    PIXEL10_12
                    PIXEL11_22
                    break;
                }
            case 56:
                {
                    PIXEL00_21
                    PIXEL01_22
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 25:
                {
                    PIXEL00_12
                    PIXEL01_22
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 26:
            case 31:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 82:
            case 214:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 88:
            case 248:
                {
                    PIXEL00_21
                    PIXEL01_22
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 74:
            case 107:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 27:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_10
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 86:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_21
                    PIXEL11_10
                    break;
                }
            case 216:
                {
                    PIXEL00_21
                    PIXEL01_22
                    PIXEL10_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 106:
                {
                    PIXEL00_10
                    PIXEL01_21
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 30:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 210:
                {
                    PIXEL00_22
                    PIXEL01_10
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 120:
                {
                    PIXEL00_21
                    PIXEL01_22
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_10
                    break;
                }
            case 75:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    PIXEL10_10
                    PIXEL11_22
                    break;
                }
            case 29:
                {
                    PIXEL00_12
                    PIXEL01_11
                    PIXEL10_22
                    PIXEL11_21
                    break;
                }
            case 198:
                {
                    PIXEL00_22
                    PIXEL01_12
                    PIXEL10_21
                    PIXEL11_11
                    break;
                }
            case 184:
                {
                    PIXEL00_21
                    PIXEL01_22
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 99:
                {
                    PIXEL00_11
                    PIXEL01_21
                    PIXEL10_12
                    PIXEL11_22
                    break;
                }
            case 57:
                {
                    PIXEL00_12
                    PIXEL01_22
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 71:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_21
                    PIXEL11_22
                    break;
                }
            case 156:
                {
                    PIXEL00_21
                    PIXEL01_11
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 226:
                {
                    PIXEL00_22
                    PIXEL01_21
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 60:
                {
                    PIXEL00_21
                    PIXEL01_11
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 195:
                {
                    PIXEL00_11
                    PIXEL01_21
                    PIXEL10_21
                    PIXEL11_11
                    break;
                }
            case 102:
                {
                    PIXEL00_22
                    PIXEL01_12
                    PIXEL10_12
                    PIXEL11_22
                    break;
                }
            case 153:
                {
                    PIXEL00_12
                    PIXEL01_22
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 58:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 83:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 92:
                {
                    PIXEL00_21
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 202:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    PIXEL01_21
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    PIXEL11_11
                    break;
                }
            case 78:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    PIXEL11_22
                    break;
                }
            case 154:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 114:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 89:
                {
                    PIXEL00_12
                    PIXEL01_22
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 90:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 55:
            case 23:
                {
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL00_11
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL00_60
                        PIXEL01_90
                    }
                    PIXEL10_20
                    PIXEL11_21
                    break;
                }
            case 182:
            case 150:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                        PIXEL11_12
                    }
                    else
                    {
                        PIXEL01_90
                        PIXEL11_61
                    }
                    PIXEL10_20
                    break;
                }
            case 213:
            case 212:
                {
                    PIXEL00_20
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL01_11
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL01_60
                        PIXEL11_90
                    }
                    PIXEL10_21
                    break;
                }
            case 241:
            case 240:
                {
                    PIXEL00_20
                    PIXEL01_22
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL10_12
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL10_61
                        PIXEL11_90
                    }
                    break;
                }
            case 236:
            case 232:
                {
                    PIXEL00_21
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                        PIXEL11_11
                    }
                    else
                    {
                        PIXEL10_90
                        PIXEL11_60
                    }
                    break;
                }
            case 109:
            case 105:
                {
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL00_12
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL00_61
                        PIXEL10_90
                    }
                    PIXEL01_20
                    PIXEL11_22
                    break;
                }
            case 171:
            case 43:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                        PIXEL10_11
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL10_60
                    }
                    PIXEL01_21
                    PIXEL11_20
                    break;
                }
            case 143:
            case 15:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                        PIXEL01_12
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL01_61
                    }
                    PIXEL10_22
                    PIXEL11_20
                    break;
                }
            case 124:
                {
                    PIXEL00_21
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_10
                    break;
                }
            case 203:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    PIXEL10_10
                    PIXEL11_11
                    break;
                }
            case 62:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 211:
                {
                    PIXEL00_11
                    PIXEL01_10
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 118:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_12
                    PIXEL11_10
                    break;
                }
            case 217:
                {
                    PIXEL00_12
                    PIXEL01_22
                    PIXEL10_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 110:
                {
                    PIXEL00_10
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 155:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_10
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 188:
                {
                    PIXEL00_21
                    PIXEL01_11
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 185:
                {
                    PIXEL00_12
                    PIXEL01_22
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 61:
                {
                    PIXEL00_12
                    PIXEL01_11
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 157:
                {
                    PIXEL00_12
                    PIXEL01_11
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 103:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_12
                    PIXEL11_22
                    break;
                }
            case 227:
                {
                    PIXEL00_11
                    PIXEL01_21
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 230:
                {
                    PIXEL00_22
                    PIXEL01_12
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 199:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_21
                    PIXEL11_11
                    break;
                }
            case 220:
                {
                    PIXEL00_21
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 158:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 234:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    PIXEL01_21
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_11
                    break;
                }
            case 242:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 59:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 121:
                {
                    PIXEL00_12
                    PIXEL01_22
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 87:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 79:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    PIXEL11_22
                    break;
                }
            case 122:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 94:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 218:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 91:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 229:
                {
                    PIXEL00_20
                    PIXEL01_20
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 167:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_20
                    PIXEL11_20
                    break;
                }
            case 173:
                {
                    PIXEL00_12
                    PIXEL01_20
                    PIXEL10_11
                    PIXEL11_20
                    break;
                }
            case 181:
                {
                    PIXEL00_20
                    PIXEL01_11
                    PIXEL10_20
                    PIXEL11_12
                    break;
                }
            case 186:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 115:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 93:
                {
                    PIXEL00_12
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 206:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    PIXEL11_11
                    break;
                }
            case 205:
            case 201:
                {
                    PIXEL00_12
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_10
                    }
                    else
                    {
                        PIXEL10_70
                    }
                    PIXEL11_11
                    break;
                }
            case 174:
            case 46:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_10
                    }
                    else
                    {
                        PIXEL00_70
                    }
                    PIXEL01_12
                    PIXEL10_11
                    PIXEL11_20
                    break;
                }
            case 179:
            case 147:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_10
                    }
                    else
                    {
                        PIXEL01_70
                    }
                    PIXEL10_20
                    PIXEL11_12
                    break;
                }
            case 117:
            case 116:
                {
                    PIXEL00_20
                    PIXEL01_11
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_10
                    }
                    else
                    {
                        PIXEL11_70
                    }
                    break;
                }
            case 189:
                {
                    PIXEL00_12
                    PIXEL01_11
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 231:
                {
                    PIXEL00_11
                    PIXEL01_12
                    PIXEL10_12
                    PIXEL11_11
                    break;
                }
            case 126:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_10
                    break;
                }
            case 219:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_10
                    PIXEL10_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 125:
                {
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL00_12
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL00_61
                        PIXEL10_90
                    }
                    PIXEL01_11
                    PIXEL11_10
                    break;
                }
            case 221:
                {
                    PIXEL00_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL01_11
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL01_60
                        PIXEL11_90
                    }
                    PIXEL10_10
                    break;
                }
            case 207:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                        PIXEL01_12
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL01_61
                    }
                    PIXEL10_10
                    PIXEL11_11
                    break;
                }
            case 238:
                {
                    PIXEL00_10
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                        PIXEL11_11
                    }
                    else
                    {
                        PIXEL10_90
                        PIXEL11_60
                    }
                    break;
                }
            case 190:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                        PIXEL11_12
                    }
                    else
                    {
                        PIXEL01_90
                        PIXEL11_61
                    }
                    PIXEL10_11
                    break;
                }
            case 187:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                        PIXEL10_11
                    }
                    else
                    {
                        PIXEL00_90
                        PIXEL10_60
                    }
                    PIXEL01_10
                    PIXEL11_12
                    break;
                }
            case 243:
                {
                    PIXEL00_11
                    PIXEL01_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL10_12
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL10_61
                        PIXEL11_90
                    }
                    break;
                }
            case 119:
                {
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL00_11
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL00_60
                        PIXEL01_90
                    }
                    PIXEL10_12
                    PIXEL11_10
                    break;
                }
            case 237:
            case 233:
                {
                    PIXEL00_12
                    PIXEL01_20
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    PIXEL11_11
                    break;
                }
            case 175:
            case 47:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    PIXEL01_12
                    PIXEL10_11
                    PIXEL11_20
                    break;
                }
            case 183:
            case 151:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_20
                    PIXEL11_12
                    break;
                }
            case 245:
            case 244:
                {
                    PIXEL00_20
                    PIXEL01_11
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 250:
                {
                    PIXEL00_10
                    PIXEL01_10
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 123:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_10
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_10
                    break;
                }
            case 95:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_10
                    PIXEL11_10
                    break;
                }
            case 222:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 252:
                {
                    PIXEL00_21
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 249:
                {
                    PIXEL00_12
                    PIXEL01_22
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 235:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_21
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    PIXEL11_11
                    break;
                }
            case 111:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_22
                    break;
                }
            case 63:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_11
                    PIXEL11_21
                    break;
                }
            case 159:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_22
                    PIXEL11_12
                    break;
                }
            case 215:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_21
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 246:
                {
                    PIXEL00_22
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 254:
                {
                    PIXEL00_10
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 253:
                {
                    PIXEL00_12
                    PIXEL01_11
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 251:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    PIXEL01_10
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 239:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    PIXEL01_12
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    PIXEL11_11
                    break;
                }
            case 127:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_20
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_20
                    }
                    PIXEL11_10
                    break;
                }
            case 191:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_11
                    PIXEL11_12
                    break;
                }
            case 223:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_20
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_10
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_20
                    }
                    break;
                }
            case 247:
                {
                    PIXEL00_11
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    PIXEL10_12
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
            case 255:
                {
                    if (patternTable[i-1] & (1 << 2)) // Diff(w[4], w[2])
                    {
                        PIXEL00_0
                    }
                    else
                    {
                        PIXEL00_100
                    }
                    if (patternTable[i+1] & (1 << 0)) // Diff(w[2], w[6])
                    {
                        PIXEL01_0
                    }
                    else
                    {
                        PIXEL01_100
                    }
                    if (patternTable[i-1] & (1 << 7)) // Diff(w[8], w[4])
                    {
                        PIXEL10_0
                    }
                    else
                    {
                        PIXEL10_100
                    }
                    if (patternTable[i+1] & (1 << 5)) // Diff(w[6], w[8])
                    {
                        PIXEL11_0
                    }
                    else
                    {
                        PIXEL11_100
                    }
                    break;
                }
        }
        rgbsrc1++;
        rgbsrc2++;
        rgbsrc3++;
        dst += 2;
    }
}

void hq2x_8_16( const uint8_t * sp, uint16_t * dp, const uint16_t * pal, int Xres, int Yres )
{
    int  j;
    uint8_t *bp;

    uint16_t *rgbsrc1;
    uint16_t *rgbsrc2;
    uint16_t *rgbsrc3;

    uint32_t *yuvsrc1;
    uint32_t *yuvsrc2;
    uint32_t *yuvsrc3;

    uint8_t *patternsrc1;

    bp = (uint8_t *) alloca(3 * sizeof(uint16_t) * Xres + 3 * sizeof(uint32_t) * Xres + 1 * sizeof(uint8_t) * Xres + (1+1+3+1) * ALIGNMENT);

    bp = (uint8_t *) ( (((uintptr_t) bp) + (ALIGNMENT - 1)) & ~((uintptr_t)(ALIGNMENT - 1)) ); // bp is aligned

    rgbsrc1 = (uint16_t *) bp; // rgbsrc1 is aligned
    rgbsrc2 = (uint16_t *) ( ((uintptr_t) rgbsrc1) + 2 * Xres); // rgbsrc2 is aligned
    rgbsrc3 = (uint16_t *) ( ((uintptr_t) rgbsrc2) + 2 * Xres); // rgbsrc3 is aligned

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) rgbsrc3) + 2 * Xres + ALIGNMENT); // yuvsrc1 is aligned and has place for yuvsrc1[-1] and yuvsrc1[Xres]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * Xres + ALIGNMENT); // yuvsrc2 is aligned and has place for yuvsrc2[-1] and yuvsrc2[Xres]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * Xres + ALIGNMENT); // yuvsrc3 is aligned and has place for yuvsrc2[-1] and yuvsrc2[Xres]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc3) + 4 * Xres + ALIGNMENT); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[Xres]


    // convert first line to yuv and rgb
    convert_yuv_rgb_8(sp, yuvsrc1, rgbsrc1, pal, Xres);

    // convert second line to yuv and rgb
    convert_yuv_rgb_8(sp + Xres, yuvsrc2, rgbsrc2, pal, Xres);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, Xres);

    fill_dst(rgbsrc1, rgbsrc1, rgbsrc2, dp, patternsrc1, Xres);

    sp += 2 * Xres;
    dp += 2*2 * Xres;

    for (j = Yres - 2; j != 0; j--)
    {
        // convert "third" line to yuv and rgb
        convert_yuv_rgb_8(sp, yuvsrc3, rgbsrc3, pal, Xres);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, Xres);

        fill_dst(rgbsrc1, rgbsrc2, rgbsrc3, dp, patternsrc1, Xres);

        sp += Xres;
        dp += 2*2 * Xres;

        // "rename" buffers
        {
            uint16_t *rgbsrc;
            rgbsrc = rgbsrc1;
            rgbsrc1 = rgbsrc2;
            rgbsrc2 = rgbsrc3;
            rgbsrc3 = rgbsrc;
        }
        {
            uint32_t *yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc;
        }
    }

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc2, patternsrc1, Xres);

    fill_dst(rgbsrc1, rgbsrc2, rgbsrc2, dp, patternsrc1, Xres);
}

void hq2x_16( const uint16_t * sp, uint16_t * dp, int Xres, int Yres )
{
    int  j;
    uint8_t *bp;

    uint32_t *yuvsrc1;
    uint32_t *yuvsrc2;
    uint32_t *yuvsrc3;

    uint8_t *patternsrc1;

    bp = (uint8_t *) alloca(3 * sizeof(uint32_t) * Xres + 1 * sizeof(uint8_t) * Xres + (1+1+3+1) * ALIGNMENT);

    bp = (uint8_t *) ( (((uintptr_t) bp) + (ALIGNMENT - 1)) & ~((uintptr_t)(ALIGNMENT - 1)) ); // bp is aligned

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) bp     )            + ALIGNMENT); // yuvsrc1 is aligned and has place for yuvsrc1[-1] and yuvsrc1[Xres]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * Xres + ALIGNMENT); // yuvsrc2 is aligned and has place for yuvsrc2[-1] and yuvsrc2[Xres]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * Xres + ALIGNMENT); // yuvsrc3 is aligned and has place for yuvsrc2[-1] and yuvsrc2[Xres]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc3) + 4 * Xres + ALIGNMENT); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[Xres]


    // convert first line to yuv
    convert_yuv_16(sp, yuvsrc1, Xres);

    // convert second line to yuv
    convert_yuv_16(sp + Xres, yuvsrc2, Xres);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, Xres);

    fill_dst(sp, sp, sp + Xres, dp, patternsrc1, Xres);

    sp += Xres;
    dp += 2*2 * Xres;

    for (j = Yres - 2; j != 0; j--)
    {
        // convert "third" line to yuv
        convert_yuv_16(sp + Xres, yuvsrc3, Xres);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, Xres);

        fill_dst(sp - Xres, sp, sp + Xres, dp, patternsrc1, Xres);

        sp += Xres;
        dp += 2*2 * Xres;

        // "rename" buffers
        {
            uint32_t *yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc;
        }
    }

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc2, patternsrc1, Xres);

    fill_dst(sp - Xres, sp, sp, dp, patternsrc1, Xres);
}

